from enum import Enum
from typing import Optional
from fakts import Config


class TransportProtocol(Enum):
    WEBSOCKET = "WEBSOCKET"
    KAFKA = "KAFKA"
    RABBITMQ = "RABBITMQ"


class PostmanSettings(Config):
    type: TransportProtocol = TransportProtocol.WEBSOCKET
    kwargs: dict = {}


class AgentSettings(Config):
    type: TransportProtocol = TransportProtocol.WEBSOCKET
    debug: bool = False
    kwargs: dict = {}


class ArkitektConfig(Config):
    host: str
    port: int
    secure: bool
    postman: Optional[PostmanSettings]
    agent: Optional[AgentSettings]

    class Config:
        group = "arkitekt"
        env_prefix = "arkitekt_"

    @property
    def endpoint(self):
        return f"http://{self.host}:{self.port}/graphql"



import asyncio
from arkitekt.config import TransportProtocol
from arkitekt.messages.base import MessageModel
from arkitekt.transport.registry import (
    TransportRegistry,
    get_current_transport_registry,
)
from arkitekt.transport.base import Transport
from arkitekt.ward import ArkitektWard
from herre.herre import Herre, get_current_herre
from herre.wards.registry import get_ward_registry
from arkitekt.legacy.utils import *
from arkitekt.registry import set_current_rpc
from koil.koil import Koil, get_current_koil
from fakts import Fakts, get_current_fakts, Config

try:
    from rich.traceback import install

    install()
except:
    pass

import logging

logger = logging.getLogger(__name__)


class PostmanConfig(Config):
    type: TransportProtocol
    kwargs: dict

    class Config:
        group = "arkitekt.postman"


class Postman:
    def __init__(
        self,
        *args,
        auto_login=True,
        auto_connect=True,
        loop=None,
        register=True,
        herre: Herre = None,
        koil: Koil = None,
        fakts: Fakts = None,
        transport_registry: TransportRegistry = None,
        **kwargs,
    ) -> None:

        self.auto_login = auto_login
        self.auto_connect = auto_connect
        self.transport: Transport = None
        self.herre = herre or get_current_herre()
        self.fakts = fakts or get_current_fakts()
        self.ward: ArkitektWard = get_ward_registry().get_ward_instance("arkitekt")
        self.transport_registry = transport_registry or get_current_transport_registry()

        self.connected = False
        self.queues = {}

        if register:
            set_current_rpc(self)

        super().__init__(*args, **kwargs)

    async def adisconnect(self):
        await self.ward.adisconnect()
        await self.transport.adisconnect()
        logger.info(f"Successfully disconnected Postman {self}")

    async def broadcast(self, message: MessageModel):
        if message.meta.reference in self.queues:
            logger.info(f"RPC Message Received")
            await self.queues[message.meta.reference].put(message)
        else:
            logger.error(f"Message without a reply channel {message}")

    async def stream_replies_to_queue(self, message: MessageModel):
        """Creates a queue for this referenced message
        and forwards every message on a reply channel to the return queue

        Args:
            message (MessageModel): [description]

        Returns:
            asyncio.Queue: The Reply Queue
        """
        reply_channel = message.meta.reference
        reply_queue = asyncio.Queue()
        self.queues[reply_channel] = reply_queue
        await self.transport.forward(message)
        return reply_queue

    async def forward(self, message: MessageModel):
        await self.transport.forward(message)

    async def aconnect(self):
        if not self.fakts.loaded:
            await self.fakts.aload()

        if not self.herre.logged_in:
            await self.herre.alogin()

        if not self.ward.connected:
            assert (
                self.auto_connect
            ), "We have not connected to Arkitekt before and autoconnect was set to false. Please connect a ward before or set auto_connect to True"
            await self.ward.aconnect()

        self.config = await PostmanConfig.from_fakts(fakts=self.fakts)
        self.transcript = self.ward.transcript
        self.transport = self.transport_registry.get_postman_transport_for_protocol(
            self.config.type
        )(self.config.kwargs, broadcast=self.broadcast)

        await self.transport.aconnect()
        self.connected = True

    async def __aenter__(self):
        await self.aconnect()
        return self

    async def __aexit__(self, *args, **kwargs):
        await self.adisconnect()



from typing import Dict
from .base import Transport
from .agent.base import AgentTransport
from .postman.base import PostmanTransport
from arkitekt.config import TransportProtocol
import logging

logger = logging.getLogger(__name__)


class TransportRegistry:
    def __init__(self) -> None:
        self.registeredAgentTransports: Dict[str, AgentTransport] = {}
        self.registeredPostmanTransports: Dict[str, PostmanTransport] = {}

    def register_agent_transport(
        self, protocol: TransportProtocol, transport: AgentTransport
    ):
        self.registeredAgentTransports[protocol] = transport

    def register_postman_transport(
        self, protocol: TransportProtocol, transport: PostmanTransport
    ):
        self.registeredPostmanTransports[protocol] = transport

    def get_postman_transport_for_protocol(self, protocol):
        return self.registeredPostmanTransports[protocol]

    def get_agent_transport_for_protocol(self, protocol):
        return self.registeredAgentTransports[protocol]


def register_agent_transport(protocol: TransportProtocol):
    def real_decorator(transport):

        assert issubclass(transport, Transport), "Transport must subclass Transport"
        logger.info(f"Registerying Agent Transport {transport} for {protocol}")
        get_current_transport_registry().register_agent_transport(protocol, transport)
        return transport

    return real_decorator


def register_postman_transport(protocol: TransportProtocol):
    def real_decorator(transport):

        assert issubclass(transport, Transport), "Transport must subclass Transport"
        logger.info(f"Registerying Postman Transport {transport} for {protocol}")
        get_current_transport_registry().register_postman_transport(protocol, transport)
        return transport

    return real_decorator


TRANSPORT_REGISTRY = None


def get_current_transport_registry(with_defaults=True):
    global TRANSPORT_REGISTRY
    if not TRANSPORT_REGISTRY:
        TRANSPORT_REGISTRY = TransportRegistry()
        if with_defaults:
            import arkitekt.transport.agent.websocket
            import arkitekt.transport.postman.websocket

    return TRANSPORT_REGISTRY



from arkitekt.messages.postman.log import LogLevel
from contextvars import ContextVar
from rich.console import ConsoleRenderable
from rich.panel import Panel
from rich.table import Table
from rich.live import Live
from herre.console.context import get_current_console

current_monitor: ContextVar["Monitor"] = ContextVar('current_monitor', default=None)


def get_current_monitor():
    global current_monitor
    return current_monitor.get()


class ReservationPanel:

    def __init__(self, monitor: "Monitor", reservation) -> None:
        self.monitor = monitor
        self.reservation = reservation

    def start(self):
        heading_information = Table.grid(expand=True)
        heading_information.add_column()
        heading_information.add_column(style="green")

        reserving_table = Table(title=f"[bold green]Reserving on ...", show_header=False)

        for key, value in self.reservation.params.dict().items():
            reserving_table.add_row(key, str(value))

        heading_information.add_row(self.reservation.node.__rich__(), reserving_table)

        self.table = Table()
        self.table.add_column("Level")
        self.table.add_column("Message")

        columns = Table.grid(expand=True)
        columns.add_column()

        columns.add_row(heading_information)
        columns.add_row(self.table)

        self.monitor.addRow(Panel(columns, title="Reservation"))

    def end(self):
        pass

    def log(self, message: str, level: LogLevel = LogLevel.INFO):
         self.table.add_row(level, f"{message}")

class ActorPanel:

    def __init__(self, monitor: "Monitor", actor) -> None:
        self.monitor = monitor
        self.actor = actor

    def start(self):
        heading_information = Table.grid(expand=True)
        heading_information.add_column()
        heading_information.add_column(style="green")

        reserving_table = Table(title=f"[bold green]Provision on ...", show_header=False)

        heading_information.add_row(self.actor.template.node.__rich__())

        self.table = Table()
        self.table.add_column("Level")
        self.table.add_column("Message")

        columns = Table.grid(expand=True)
        columns.add_column()

        columns.add_row(heading_information)
        columns.add_row(self.table)

        self.monitor.addRow(Panel(columns, title="Provision"))

    def end(self):
        pass

    def log(self, message: str, level: LogLevel = LogLevel.INFO):
        self.table.add_row(level, f"{message}")


class AgentPanel:

    def __init__(self, monitor: "Monitor", provision) -> None:
        self.monitor = monitor
        self.provision = provision

    def start(self):
        heading_information = Table.grid(expand=True)
        heading_information.add_column()
        heading_information.add_column(style="green")


        self.table = Table()
        self.table.add_column("Template")
        self.table.add_column("Agent")

        columns = Table.grid(expand=True)
        columns.add_column()

        columns.add_row(heading_information)
        columns.add_row(self.table)

        self.monitor.addRow(Panel(columns, title="Agent Facts"))

    def end(self):
        pass

    def add_to_actor_map(self, template, actor):
        self.table.add_row(f"{template.id} - {template.node.name}", f"{actor.__name__}")



class Monitor:

    def __init__(self, title="Monitor", log=False) -> None:
        """Monitor allows you to monitor the progress of what is happenening inside your application


        Args:
            title (str, optional): The Title of this Monitor (Is the Panels Title). Defaults to "Monitor".
            progress (bool, optional): Do you want to monitor the progress of assignments and reservations? Defaults to False.
        """
        self.columns = Table.grid(expand=True)
        self.columns.add_column()
        self.log = log
        self.panel = Panel(self.columns, title=title)
        self.live = Live(self.panel, refresh_per_second=4, console=get_current_console())


    def create_reservation_panel(self, reservation):
        return ReservationPanel(self, reservation)

    def create_actor_panel(self, provision):
        return ActorPanel(self, provision)

    def create_agent_panel(self, agent):
        return AgentPanel(self, agent)

    def addRow(self, renderable: ConsoleRenderable):
        self.columns.add_row(renderable)

    def __aenter__(self):
        '''Convenience Method'''
        return self.__aenter__()

    def __aexit__(self,*args,**kwargs):
        self.__exit__(*args,**kwargs)

    def __enter__(self):
        current_monitor.set(self)
        self.live.__enter__()
        return self


    def __exit__(self, *args, **kwargs):
        self.live.__exit__(*args, **kwargs)
        current_monitor.set(None)



from arkitekt.agents.base import Agent
from arkitekt.messages.postman.provide.provide_log import ProvideLogMessage
from arkitekt.messages.postman.log import LogLevel
from arkitekt.monitor.monitor import AgentPanel, get_current_monitor
from herre.console.context import get_current_console
from arkitekt.packers.utils import shrink_outputs
from re import template
from arkitekt.messages.postman.assign.assign_cancelled import AssignCancelledMessage
from arkitekt.messages.postman.unassign.bounced_forwarded_unassign import (
    BouncedForwardedUnassignMessage,
)
from arkitekt.messages.postman.assign.assign_critical import AssignCriticalMessage
from typing import Dict
from arkitekt.messages.postman.assign.assign_return import AssignReturnMessage
from arkitekt.messages.postman.assign.bounced_forwarded_assign import (
    BouncedForwardedAssignMessage,
)
from arkitekt.messages.postman.assign.bounced_assign import BouncedAssignMessage
from arkitekt.messages.base import MessageModel
from arkitekt.messages.postman.provide.provide_transition import (
    ProvideMode,
    ProvideState,
)
from arkitekt.transport.base import Transport
from arkitekt.packers.transpilers.base import Transpiler
from arkitekt.messages.postman.provide.bounced_provide import BouncedProvideMessage
import asyncio
from asyncio.tasks import Task, create_task
from arkitekt.messages.postman.provide import ProvideTransitionMessage
import logging
from arkitekt.schema.template import Template
from koil.koil import Koil, get_current_koil
from koil.loop import koil


logger = logging.getLogger(__name__)


class Actor:
    template: Template

    def __init__(
        self,
        *args,
        koil: Koil = None,
        strict=False,
        expand_inputs=True,
        shrink_outputs=True,
        transpilers={},
        **kwargs,
    ) -> None:
        super().__init__(*args, **kwargs)
        self.strict = strict
        self.expand_inputs = expand_inputs
        self.shrink_outputs = shrink_outputs
        self.runningAssignments: Dict[
            str, asyncio.Task
        ] = {}  # Running assignments indexed by assignment reference

    async def log(self, message, level=LogLevel.INFO):
        logger.info("{level}: {message}")

    async def provide_log(self, message: str, level=LogLevel.INFO):
        await self.agent.transport.forward(
            ProvideLogMessage(
                data={"level": level, "message": message},
                meta=self.provision.meta.dict(exclude={"type"}),
            )
        )
        logger.info(f"Provide Log: {level}: {message}")

    def run(self, *args, **kwargs):
        return koil(self.arun(*args, **kwargs))

    async def acall(self, message: MessageModel):
        await self.in_queue.put(message)

    async def on_provide(self, message: BouncedProvideMessage):
        return None

    async def on_unprovide(self, message: BouncedProvideMessage):
        return None

    async def on_assign(self, message: BouncedForwardedAssignMessage):
        raise NotImplementedError("Needs to be owerwritten in Actor Subclass")

    async def arun(self, provision: BouncedProvideMessage, agent: Agent):
        self.loop = asyncio.get_running_loop()
        self.provision = provision
        self.agent = agent
        self.transport = agent.transport
        self.in_queue = asyncio.Queue()

        try:
            self.template = await Template.asyncs.get(id=self.provision.data.template)

            await self.transport.forward(
                ProvideTransitionMessage(
                    data={
                        "state": ProvideState.PROVIDING,
                        "message": "We just got started Bay",
                    },
                    meta={
                        "reference": self.provision.meta.reference,
                        "extensions": self.provision.meta.extensions,
                    },
                )
            )

            await self.on_provide(self.provision)

            await self.transport.forward(
                ProvideTransitionMessage(
                    data={
                        "state": ProvideState.ACTIVE,
                        "message": "We just got started Bay",
                        "mode": ProvideMode.DEBUG
                        if agent.config.debug
                        else ProvideMode.PRODUCTION,
                    },
                    meta={
                        "reference": self.provision.meta.reference,
                        "extensions": self.provision.meta.extensions,
                    },
                )
            )

            while True:
                await self.log("Waiting for assignmements")
                message = await self.in_queue.get()
                logger.info(f"Received Message {message}")

                if isinstance(message, BouncedForwardedAssignMessage):
                    await self.log("Assigningment received")
                    task = create_task(self.on_assign(message))
                    self.runningAssignments[message.meta.reference] = task

                if isinstance(message, BouncedForwardedUnassignMessage):
                    if message.data.assignation in self.runningAssignments:
                        task = self.runningAssignments[message.data.assignation]
                        if not task.done():
                            logger.info("Task is being cancelled")
                            task.cancel()
                        else:
                            logger.error("Task was already done")
                    else:
                        logger.error("Task was never assigned to this actor")
                        if self.strict:
                            raise Exception(
                                "Received cancellation for Task that was never assinged to this actor!"
                            )
                        await self.transport.forward(
                            AssignCancelledMessage(
                                data={
                                    "canceller": str(
                                        "Cancelled because actor receiving this cancellation never had this task but was also not strict"
                                    )
                                },
                                meta={
                                    "reference": message.data.assignation,
                                    "extensions": message.meta.extensions,
                                },
                            )
                        )

        except Exception as e:
            logger.exception(e)
            await self.log(f"Provision Exception {str(e)}")
            await self.transport.forward(
                ProvideTransitionMessage(
                    data={"state": ProvideState.CRITICAL, "message": f"{e}"},
                    meta={
                        "reference": self.provision.meta.reference,
                        "extensions": self.provision.meta.extensions,
                    },
                )
            )

        except asyncio.CancelledError as e:
            await self.transport.forward(
                ProvideTransitionMessage(
                    data={"state": ProvideState.CANCELING, "message": f"{e}"},
                    meta={
                        "reference": self.provision.meta.reference,
                        "extensions": self.provision.meta.extensions,
                    },
                )
            )

            await self.on_unprovide(self.provision)

            logger.info("Doing Whatever needs to be done to cancel!")
            await self.transport.forward(
                ProvideTransitionMessage(
                    data={"state": ProvideState.CANCELLED, "message": f"{e}"},
                    meta={
                        "reference": self.provision.meta.reference,
                        "extensions": self.provision.meta.extensions,
                    },
                )
            )
            raise e



from pydantic.main import BaseModel
from websockets.exceptions import ConnectionClosedError
from arkitekt.config import TransportProtocol
from arkitekt.transport.agent.base import AgentTransportConfig
from arkitekt.transport.base import Transport, TransportConfig
from arkitekt.messages.utils import expandToMessage
from arkitekt.messages.base import MessageMetaExtensionsModel, MessageModel
from arkitekt.transport.registry import register_agent_transport
from herre.herre import get_current_herre
from enum import Enum
import websockets
import json
import asyncio
from arkitekt.legacy.utils import create_task
import logging

logger = logging.getLogger(__name__)


class AgentWebsocketCodes(int, Enum):
    ALREADY_CONNECTED_IDENTIFER = 4001


class WebsocketAgentTransportConfig(AgentTransportConfig):
    host: str
    port: int
    secure: bool = False

    @property
    def protocol(self):
        return "wss" if self.secure else "ws"


class ConnectionFailedError(Exception):
    pass


class CorrectableConnectionFail(ConnectionFailedError):
    pass


class DefiniteConnectionFail(ConnectionFailedError):
    pass


@register_agent_transport(TransportProtocol.WEBSOCKET)
class WebsocketAgentTransport(Transport):
    configClass = WebsocketAgentTransportConfig
    config: WebsocketAgentTransportConfig

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.send_queue = None
        self.retries = 5
        self.time_between_retries = 5
        self.connection_alive = False
        self.connection_dead = False

    async def aconnect(self):
        if not self.herre.logged_in:
            await self.herre.alogin()

        self.send_queue = asyncio.Queue()
        await self.websocket_loop()

    async def adisconnect(self):
        logger.info(f"Websocket Transport {self} succesfully disconnected")

    async def websocket_loop(self, retry=0):
        send_task = None
        receive_task = None
        try:
            try:
                async with websockets.connect(
                    f"{self.config.protocol}://{self.config.host}:{self.config.port}/agent/?token={self.herre.state.access_token}&identifier={self.config.identifier}"
                ) as client:

                    send_task = create_task(self.sending(client))
                    receive_task = create_task(self.receiving(client))

                    self.connection_alive = True
                    self.connection_dead = False
                    done, pending = await asyncio.wait(
                        [send_task, receive_task],
                        return_when=asyncio.FIRST_EXCEPTION,
                    )
                    self.connection_alive = True

                    for task in pending:
                        task.cancel()

                    for task in done:
                        raise task.exception()

            except ConnectionClosedError as e:
                logger.exception(e)
                if e.code == AgentWebsocketCodes.ALREADY_CONNECTED_IDENTIFER:
                    raise DefiniteConnectionFail(
                        f"Another agent was already connected on {self.config.identifier}! Please set a different identifier if you want to connect!"
                    ) from e

                raise CorrectableConnectionFail from e

            except Exception as e:
                logger.warning(
                    "THIS EXCEPTION HAS NO RETRY STRATEGY... TRYING TO RETRY??"
                )
                raise CorrectableConnectionFail from e

        except CorrectableConnectionFail as e:
            logger.info(f"Trying to Recover from Exception {e}")
            if retry > self.retries:
                raise DefiniteConnectionFail("Exceeded Number of Retries")
            await asyncio.sleep(self.time_between_retries)
            logger.info(f"Retrying to connect")
            await self.websocket_loop(retry=retry + 1)

        except DefiniteConnectionFail as e:
            self.connection_dead = False
            raise e

        except asyncio.CancelledError as e:
            logger.info("Got Canceleld")
            if send_task and receive_task:
                send_task.cancel()
                receive_task.cancel()

            cancellation = await asyncio.gather(
                send_task, receive_task, return_exceptions=True
            )
            raise e

    async def sending(self, client):
        try:
            while True:
                message = await self.send_queue.get()
                logger.debug("Agent Websocket: >>>>>> " + message)
                await client.send(message)
                self.send_queue.task_done()
        except asyncio.CancelledError as e:
            logger.debug("Sending Task sucessfully Cancelled")

    async def receiving(self, client):
        try:
            async for message in client:
                logger.debug("Agent Websocket: <<<<<<< " + message)
                message = expandToMessage(json.loads(message))
                await self.broadcast(message)
        except asyncio.CancelledError as e:
            logger.debug("Receiving Task sucessfully Cancelled")

    async def forward(self, message: MessageModel):
        assert (
            not self.connection_dead
        ), "Connection is definetly dead. Retries have been exceeded. Error"
        await self.send_queue.put(json.dumps(message.dict()))



from ....messages.generics import Context
from pydantic.main import BaseModel
from ....messages.types import BOUNCED_ASSIGN, BOUNCED_UNASSIGN
from ....messages.base import (
    MessageDataModel,
    MessageMetaExtensionsModel,
    MessageMetaModel,
    MessageModel,
)
from typing import List, Optional


class AssignParams(BaseModel):
    providers: Optional[List[str]]


class MetaExtensionsModel(MessageMetaExtensionsModel):
    # Set by postman consumer
    progress: Optional[str]
    callback: Optional[str]


class MetaModel(MessageMetaModel):
    type: str = BOUNCED_UNASSIGN
    extensions: Optional[MetaExtensionsModel]
    context: Context


class DataModel(MessageDataModel):
    assignation: str
    provision: str


class BouncedUnassignMessage(MessageModel):
    data: DataModel
    meta: MetaModel



from arkitekt.graphql.admin import RESET_REPOSITORY
from arkitekt.schema import Node
import asyncio
from herre.wards.registry import get_ward_registry
from koil.loop import koil
import re

package_test = re.compile(r"@(?P<package>[^\/]*)\/(?P<interface>[^\/]*)")

async def ause(package=None, interface=None, q=False) -> Node:
    """Use a Node on the Platform by Searching for it on its package

    Args:
        package ([type], optional): The package this Node belongs to. Defaults to None.
        interface ([type], optional):  The interface of this Node. Defaults to None.

    Returns:
        Node: The Node
        
    """
    return await Node.asyncs.get(package=package, interface=interface ,q = q)


async def areset_repository(**kwargs) -> Node:
    arkitekt_ward = get_ward_registry().get_ward_instance("arkitekt")
    return await arkitekt_ward.arun(RESET_REPOSITORY)


def use(**kwargs) -> Node:
    """Use a Node on the Platform by Searching for it on its package

    Args:
        package ([type], optional): The package this Node belongs to. Defaults to None.
        interface ([type], optional):  The interface of this Node. Defaults to None.

    Returns:
        Node: The Node
        
    """
    return koil(ause(**kwargs))


def reset_repository(**kwargs) -> Node:
    """Use a Node on the Platform by Searching for it on its package

    Args:
        package ([type], optional): The package this Node belongs to. Defaults to None.
        interface ([type], optional):  The interface of this Node. Defaults to None.

    Returns:
        Node: The Node
        
    """
    return koil(areset_repository(**kwargs))






from abc import abstractmethod
from enum import Enum
import inspect
from arkitekt.packers.models.base import StructureModel
from arkitekt.schema.widgets import AllWidgets, QueryWidget, SearchWidget, SliderWidget
from arkitekt.packers.structure import BoundType, Structure
from typing import Any, Dict, List, Optional, Union
from pydantic import BaseModel, Field, validator
from herre.access.object import GraphQLObject
from arkitekt.packers.registry import get_packer_registry
from typing import ForwardRef
import asyncio


class Port(GraphQLObject):
    key: Optional[str]
    description: Optional[str]
    label: Optional[str]

    @classmethod
    def from_params(cls, **kwargs):
        # TODO: This weird widget conversion thing needs to stop, type GraphQLOBject correctly
        port = cls(
            __typename=cls.__name__, **kwargs
        )  # We ensure creation of a proper object)
        return port


class ArgPort(Port):
    widget: Optional[AllWidgets]

    def _repr_html_list(self):
        nana = f"""
        <div class="container" style="border:1px solid #00000f;padding: 4px;">
            <div class="item item-1 font-xl">{self.key}</div>
            <div class="item item-2">{self.description}</div>
        """
        if self.widget:
            nana += self.widget._repr_html_list()
        return nana + "</div>"


class IntExpandShrink:
    async def expand(self, value, **kwargs):
        return int(value) if value is not None else getattr(self, "defaultInt", None)

    async def shrink(self, instance, **kwargs):
        return int(instance) if instance is not None else None

    def to_type(self):
        return int


class BoolExpandShrink:
    async def expand(self, value, **kwargs):
        return bool(value) if value is not None else getattr(self, "defaultBool", None)

    async def shrink(self, instance, **kwargs):
        return bool(instance) if instance is not None else None

    def to_type(self):
        return bool


class EnumExpandShrink:
    async def expand(self, value, **kwargs):
        return (
            next(
                (
                    key
                    for key, optionvalue in self.options.items()
                    if optionvalue == value
                ),
                None,
            )
            if value is not None
            else getattr(self, "defaultEnum", None)
        )

    async def shrink(self, instance, **kwargs):
        return self.options[instance] if instance else None

    def to_type(self):
        return Enum


class StringExpandShrink:
    async def expand(self, value, **kwargs):
        return str(value) if value is not None else getattr(self, "defaultString", None)

    async def shrink(self, instance, **kwargs):
        return str(instance) if instance is not None else None

    def to_type(self):
        return str


class StructureExpandShrink:
    @classmethod
    def from_structure(cls, structure: Structure, **overwrites):
        meta = structure.get_structure_meta()
        return cls.from_params(**{**meta.dict(), **overwrites})

    async def expand(self, value, transpile=True):
        if value is None:
            value = getattr(self, "defaultID", None)

        if value is None:
            return None

        structure = get_packer_registry().get_structure(self.identifier)
        return await structure.expand(value)

    async def shrink(self, instance, **kwargs):
        if instance is None:
            return None

        if hasattr(instance, "shrink"):
            return await instance.shrink()
        # Instance we are trying to shrink needs to be transpile to the required model
        from arkitekt.packers.transpilers.registry import get_transpiler_registry

        transpiler = get_transpiler_registry().get_transpiler(
            instance.__class__.__name__, self.identifier
        )
        transpiled_instance = await transpiler.transpile(instance)
        return await transpiled_instance.shrink()

    def to_type(self):
        return get_packer_registry().get_structure(self.identifier)


class ListExpandShrink:
    async def expand(self, value, **kwargs):
        value = getattr(self, "defaultList", None)

        return (
            await asyncio.gather(*[self.child.expand(item, **kwargs) for item in value])
            if value is not None
            else None
        )

    async def shrink(self, instance, **kwargs):
        assert isinstance(
            instance, list
        ), f"ListPorts only accept lists! Got {instance}"

        return (
            await asyncio.gather(
                *[self.child.shrink(item, **kwargs) for item in instance]
            )
            if instance is not None
            else None
        )

    def to_type(self):
        return List[self.child.to_type()]


class DictExpandShrink:
    async def expand(self, value, **kwargs):
        value = getattr(self, "defaultList", None)

        return (
            {
                key: await self.child.expand(item, **kwargs)
                for key, item in value.items()
            }
            if value is not None
            else None
        )

    async def shrink(self, instance, **kwargs):
        return (
            {
                key: await self.child.shrink(item, **kwargs)
                for key, item in instance.items()
            }
            if instance is not None
            else None
        )

    def to_type(self):
        return Dict[self.child.to_type()]


# Args
class EnumArgPort(ArgPort, EnumExpandShrink):
    options: Optional[dict]
    pass


class IntArgPort(ArgPort, IntExpandShrink):
    pass


class BoolArgPort(ArgPort, BoolExpandShrink):
    pass


class StringArgPort(ArgPort, StringExpandShrink):
    pass


class StructureArgPort(ArgPort, StructureExpandShrink):
    identifier: str
    bound: BoundType = BoundType.GLOBAL


ListArgPort = ForwardRef("ListArgPort")


class ListArgPort(ListExpandShrink, ArgPort):
    child: Union[IntArgPort, StructureArgPort, StringArgPort, ListArgPort]


ListArgPort.update_forward_refs()

DictArgPort = ForwardRef("DictArgPort")


class DictArgPort(DictExpandShrink, ArgPort):
    child: Union[IntArgPort, StructureArgPort, StringArgPort, ListArgPort]


DictArgPort.update_forward_refs()
# Kwargs


class KwargPort(Port):
    widget: Optional[AllWidgets]


class IntKwargPort(KwargPort, IntExpandShrink):
    defaultInt: Optional[int]

    @classmethod
    def from_params(cls, widget=None, default=None, **kwargs):
        # TODO: This weird widget conversion thing needs to stop, type GraphQLOBject correctly
        port = cls(
            __typename=cls.__name__,
            widget=widget.dict() if widget else None,
            defaultInt=default,
            **kwargs,
        )  # We ensure creation of a proper object)
        return port


class BoolKwargPort(KwargPort, BoolExpandShrink):
    defaultBool: Optional[bool]

    @classmethod
    def from_params(cls, widget=None, default=None, **kwargs):
        # TODO: This weird widget conversion thing needs to stop, type GraphQLOBject correctly
        port = cls(
            __typename=cls.__name__,
            widget=widget.dict() if widget else None,
            defaultBool=default,
            **kwargs,
        )  # We ensure creation of a proper object)
        return port


class EnumKwargPort(KwargPort, EnumExpandShrink):
    defaultValue: Any
    options: Optional[dict]
    pass

    @classmethod
    def from_params(cls, widget=None, default=None, **kwargs):
        port = cls(
            __typename=cls.__name__,
            widget=widget.dict() if widget else None,
            defaultValue=default._value_,
            **kwargs,
        )  # We ensure creation of a proper object)
        return port


class StringKwargPort(KwargPort, StringExpandShrink):
    defaultString: Optional[str]

    @classmethod
    def from_params(cls, widget=None, default=None, **kwargs):
        # TODO: This weird widget conversion thing needs to stop, type GraphQLOBject correctly
        port = cls(
            __typename=cls.__name__,
            widget=widget.dict() if widget else None,
            defaultString=default,
            **kwargs,
        )  # We ensure creation of a proper object)
        return port


class StructureKwargPort(KwargPort, StructureExpandShrink):
    defaultID: Optional[str]
    identifier: str

    @classmethod
    def from_params(cls, widget=None, default=None, **kwargs):
        # TODO: This weird widget conversion thing needs to stop, type GraphQLOBject correctly
        port = cls(
            __typename=cls.__name__,
            widget=widget.dict() if widget else None,
            defaultID=default,
            **kwargs,
        )  # We ensure creation of a proper object)
        return port


ListKwargPort = ForwardRef("ListKwargPort")


class ListKwargPort(KwargPort, ListExpandShrink):
    defaultList: Optional[List]
    child: Union[IntKwargPort, StructureKwargPort, StringKwargPort, ListKwargPort]

    @classmethod
    def from_params(cls, widget=None, default=None, **kwargs):
        # TODO: This weird widget conversion thing needs to stop, type GraphQLOBject correctly
        port = cls(
            __typename=cls.__name__,
            widget=widget.dict() if widget else None,
            defaultList=default,
            **kwargs,
        )  # We ensure creation of a proper object)
        return port


ListKwargPort.update_forward_refs()

DictKwargPort = ForwardRef("DictKwargPort")


class DictKwargPort(KwargPort, DictExpandShrink):
    defaultDict: Optional[Dict]
    child: Union[
        IntKwargPort, StructureKwargPort, StringKwargPort, ListKwargPort, DictKwargPort
    ]

    @classmethod
    def from_params(cls, widget=None, default=None, **kwargs):
        # TODO: This weird widget conversion thing needs to stop, type GraphQLOBject correctly
        port = cls(
            __typename=cls.__name__,
            widget=widget.dict() if widget else None,
            defaultDict=default,
            **kwargs,
        )  # We ensure creation of a proper object)
        return port


DictKwargPort.update_forward_refs()


# Returns


class ReturnPort(Port):
    pass


class StructureReturnPort(ReturnPort, StructureExpandShrink):
    identifier: str


class IntReturnPort(ReturnPort, IntExpandShrink):
    pass


class BoolReturnPort(ReturnPort, BoolExpandShrink):
    pass


class StringReturnPort(ReturnPort, StringExpandShrink):
    pass


ListReturnPort = ForwardRef("ListReturnPort")


class ListReturnPort(ReturnPort, ListExpandShrink):
    child: Union[IntReturnPort, StructureReturnPort, StringReturnPort, ListReturnPort]


ListReturnPort.update_forward_refs()


DictReturnPort = ForwardRef("DictReturnPort")


class DictReturnPort(ReturnPort, DictExpandShrink):
    child: Union[
        IntReturnPort,
        StructureReturnPort,
        StringReturnPort,
        ListReturnPort,
        DictReturnPort,
    ]


DictReturnPort.update_forward_refs()


AllArgPort = Union[
    StructureArgPort,
    StringArgPort,
    IntArgPort,
    ListArgPort,
    DictArgPort,
    EnumArgPort,
    BoolArgPort,
]
AllKwargPort = Union[
    StructureKwargPort,
    StringKwargPort,
    IntKwargPort,
    ListKwargPort,
    DictKwargPort,
    EnumKwargPort,
    BoolKwargPort,
]
AllReturnPort = Union[
    StructureReturnPort,
    StringReturnPort,
    IntReturnPort,
    ListReturnPort,
    DictReturnPort,
    BoolReturnPort,
]



from abc import abstractmethod
from contextvars import Context
from arkitekt.messages.postman.reserve.reserve_transition import ReserveState
from arkitekt.monitor.monitor import Monitor
from arkitekt.schema.enums import NodeType
from arkitekt.schema.ports import AllArgPort, AllKwargPort, AllReturnPort
from arkitekt.graphql.node import NODE_CREATE_QUERY, NODE_GET_QUERY
from enum import Enum
from typing import Any, List, Optional, Union
from arkitekt.contracts.reservation import Reservation
from arkitekt.schema.basic import AllRepository, Repository
from herre.access.model import GraphQLModel
from rich.table import Table

from koil.loop import koil, koil_gen


class Node(GraphQLModel):
    name: Optional[str]
    description: Optional[str]
    package: Optional[str]
    repository: Optional[AllRepository]
    interface: Optional[str]
    args: Optional[List[AllArgPort]]
    kwargs: Optional[List[AllKwargPort]]
    returns: Optional[List[AllReturnPort]]
    type: Optional[NodeType]

    def reserve(
        self,
        reference: str = None,
        provision: str = None,
        monitor: Monitor = None,
        ignore_node_exceptions=False,
        transition_hook=None,
        with_log=False,
        enter_on=[ReserveState.ACTIVE],
        exit_on=[ReserveState.ERROR, ReserveState.CANCELLED, ReserveState.CRITICAL],
        context: Context = None,
        loop=None,
        **params,
    ) -> Reservation:
        return Reservation(
            self,
            reference=reference,
            provision=provision,
            monitor=monitor,
            transition_hook=transition_hook,
            with_log=with_log,
            enter_on=enter_on,
            ignore_node_exceptions=ignore_node_exceptions,
            exit_on=exit_on,
            context=context,
            loop=loop,
            **params,
        )

    async def call_async_func(self, *args, reserve_params={}, **kwargs):
        async with self.reserve(**reserve_params) as res:
            return await res.assign_async(*args, **kwargs)

    async def call_async_gen(self, *args, reserve_params={}, **kwargs):
        async with self.reserve(**reserve_params) as res:
            async for result in res.stream_async(*args, **kwargs):
                yield result

    def __call__(self, *args: Any, fill_graphical=True, **kwargs) -> Any:
        """Call

        Call is a convenience on max function, its reserves the Node and wraps it either as
        an geneator (both async and non async depending on context) or call it as a function
        this should only be done if you know what you are doing.

        Args:
            reserve_params (dict, optional): [description]. Defaults to {}.

        Returns:
            Any: Generator or Function
        """
        if self.type == NodeType.FUNCTION:
            return koil(self.call_async_func(*args, **kwargs))

        if self.type == NodeType.GENERATOR:
            return koil_gen(self.call_async_gen(*args, **kwargs))

    class Meta:
        register = False
        ward = "arkitekt"
        get = NODE_GET_QUERY
        create = NODE_CREATE_QUERY

    def __rich_repr__(self):
        yield self.name
        yield "args", self.args
        yield "kwargs", self.kwargs
        yield "returns", self.returns

    def __rich__(self):
        my_table = Table(title=f"Node: {self.name}", show_header=False)

        my_table.add_row("ID", str(self.id))
        my_table.add_row("Package", self.package)
        my_table.add_row("Interface", self.interface)
        my_table.add_row("Type", self.type)

        return my_table

    def _repr_html_(self):
        return f"""
        <div class="container" style="border:1px solid #00000f;padding: 4px;">
            <div class="item item-1 font-xl">{self.name}</div>
            <div class="item item-2">{self.package}/{self.interface}</div>
            <div class="item item-3">Args: {" ".join([port._repr_html_list() for port in self.args])}</div>
            <div class="item item-3">Kwargs: {" ".join([port.key for port in self.kwargs])}</div>
            <div class="item item-3">Returns: {" ".join([port.key for port in self.returns])}</div>
        </div>"""



from arkitekt.graphql.template import UPDATE_OR_CREATE_TEMPLATE, TEMPLATE_GET_QUERY
from typing import Any, List, Optional, Union
from arkitekt.schema.basic import Registry

from herre.access.model import GraphQLModel
from arkitekt.schema.node import Node
from herre.wards.graphql import ParsedQuery


class Template(GraphQLModel):
    node: Optional[Node]
    registry: Optional[Registry]
    channel: Optional[str]
    version: Optional[str]

    class Meta:
        register = False
        ward = "arkitekt"
        create = UPDATE_OR_CREATE_TEMPLATE
        get = TEMPLATE_GET_QUERY



try:
    from asyncio import create_task

    create_task = create_task
except:
    from asyncio import ensure_future

    create_task = ensure_future

try:
    from asyncio import get_running_loop

    get_running_loop = get_running_loop
except:
    from asyncio import get_event_loop

    get_running_loop = get_event_loop




from typing import Dict
from arkitekt.actors.actify import actify, define
from arkitekt.agents.app import AppAgent
from arkitekt.qt.actor import QtActor
from qtpy.QtCore import QObject, Signal
from arkitekt.messages.postman.assign.assign_cancelled import AssignCancelledMessage
from arkitekt.messages.postman.unassign.bounced_forwarded_unassign import (
    BouncedForwardedUnassignMessage,
)
from arkitekt.messages.postman.provide.provide_transition import (
    ProvideState,
    ProvideTransitionMessage,
)
from arkitekt.messages.postman.assign.assign_log import AssignLogMessage
from arkitekt.messages.postman.assign.assign_critical import AssignCriticalMessage
from arkitekt.messages.postman.assign.bounced_forwarded_assign import (
    BouncedForwardedAssignMessage,
)
from arkitekt.messages.postman.log import LogLevel
from arkitekt.messages.postman.provide.provide_log import ProvideLogMessage
import asyncio
from arkitekt.actors.base import Actor
from arkitekt.messages.postman.unprovide.bounced_unprovide import (
    BouncedUnprovideMessage,
)
from arkitekt.messages.postman.provide.bounced_provide import BouncedProvideMessage
from arkitekt.schema.node import Node
from arkitekt.packers.transpilers import Transpiler
from arkitekt.agents.base import Agent, AgentException
import logging
import uuid

logger = logging.getLogger(__name__)


class AgentSignals(QObject):
    provide = Signal(BouncedProvideMessage)
    unprovide = Signal(BouncedUnprovideMessage)
    provide_transition = Signal(ProvideTransitionMessage)

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)


class QtAgent(AppAgent, QObject):
    provide_signal = Signal(bool)
    provision_signal = Signal(BouncedProvideMessage)
    unprovision_signal = Signal(BouncedUnprovideMessage)

    ACTOR_PENDING_MESSAGE = "Actor is Pending"

    def __init__(self, *args, strict=False, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.strict = strict
        self.assignFutures = {}
        self.provideFutures = {}
        self.unprovideFutures = {}
        self.appWorkers = {}

    async def on_bounced_provide(self, message: BouncedProvideMessage):
        self.provision_signal.emit(message)
        return await super().on_bounced_provide(message)

    async def on_bounced_unprovide(self, message: BouncedUnprovideMessage):
        self.unprovision_signal.emit(message)
        return await super().on_bounced_provide(message)

    def register_ui(
        self,
        function_query_or_node,
        widgets={},
        transpilers: Dict[str, Transpiler] = None,
        on_provide=None,
        on_unprovide=None,
        on_assign=None,
        timeout=500,
        **params
    ) -> QtActor:

        # Simple bypass for now
        defined_actor = QtActor(
            qt_assign=on_assign,
            qt_on_provide=on_provide,
            qt_on_unprovide=on_unprovide,
            timeout=timeout,
            **params
        )

        actor_builder = lambda: defined_actor

        if isinstance(function_query_or_node, str):
            self.templatedUnqueriedNodes.append(
                ({"q": function_query_or_node}, actor_builder, params)
            )

        if isinstance(function_query_or_node, Node):
            self.templatedNodes.append((function_query_or_node, actor_builder, params))

        else:
            defined_node = define(function=function_query_or_node, widgets=widgets)
            self.templatedNewNodes.append((defined_node, actor_builder, params))

        return defined_actor

    def register_side(
        self,
        *args,
        widgets={},
        transpilers: Dict[str, Transpiler] = None,
        on_provide=None,
        on_unprovide=None,
        on_assign=None,
        timeout=500,
        **params
    ):

        # Simple bypass for now

        if len(args) == 0:
            raise NotImplementedError(
                "Please provide either a function to create a node or a function and a node or query as arguments"
            )

        defined_actor = actify(
            args[0], on_provide=on_provide, on_unprovide=on_unprovide, **params
        )

        if len(args) == 1:
            new_node = define(function=args[0], widgets=widgets)
            self.templatedNewNodes.append((new_node, defined_actor, params))

        if len(args) == 2:
            query_or_node = args[1]

            if isinstance(query_or_node, str):
                self.templatedUnqueriedNodes.append(
                    ({"q": query_or_node}, defined_actor, params)
                )

            if isinstance(query_or_node, Node):
                self.templatedNodes.append((query_or_node, defined_actor, params))

        return defined_actor

    async def approve_nodes_and_templates(self):
        await super().approve_nodes_and_templates()
        self.provide_signal.emit(True)
        return

    async def aprovide(self):
        try:
            await super().aprovide()
        except asyncio.CancelledError as e:
            self.provide_signal.emit(False)
            raise e



from typing import Optional, Union
from herre.access.model import GraphQLModel
from herre.types import User, App


class Repository(GraphQLModel):
    name: Optional[str]

    class Meta:
        register = False
        ward = "arkitekt"


class AppRepository(Repository):
    app: Optional[App]

    class Meta:
        register = True
        ward = "arkitekt"


class MirrorRepository(Repository):
    url: Optional[str]

    class Meta:
        register = True
        ward = "arkitekt"


AllRepository = Union[MirrorRepository, AppRepository]


class Registry(GraphQLModel):
    app: Optional[App]
    user: Optional[User]

    class Meta:
        register = False
        ward = "arkitekt"



from abc import abstractmethod
from arkitekt.config import TransportProtocol
from arkitekt.messages.postman.assign.bounced_forwarded_assign import (
    BouncedForwardedAssignMessage,
)
from arkitekt.messages.postman.unassign.bounced_forwarded_unassign import (
    BouncedForwardedUnassignMessage,
)
from arkitekt.monitor.monitor import Monitor
from arkitekt.messages.postman.unprovide.bounced_unprovide import (
    BouncedUnprovideMessage,
)
from arkitekt.messages.postman.provide.bounced_provide import BouncedProvideMessage
from arkitekt.messages.postman.log import LogLevel
from arkitekt.messages.postman.reserve.bounced_reserve import BouncedReserveMessage
from arkitekt.transport.agent.base import AgentTransport
from arkitekt.transport.base import Transport
from arkitekt.messages.base import MessageDataModel, MessageModel
import asyncio
from typing import Callable, Dict, List, Tuple, Type
from arkitekt.transport.registry import (
    TransportRegistry,
    get_current_transport_registry,
)
from arkitekt.schema.node import Node
from arkitekt.schema.template import Template
from arkitekt.ward import ArkitektConfig, ArkitektWard
from arkitekt.messages import ProvideLogMessage, ProvideCriticalMessage
from herre.console.context import get_current_console
from herre.herre import Herre, get_current_herre
from herre.wards.registry import get_ward_registry
from herre.wards.base import WardException
from arkitekt.legacy.utils import *
from arkitekt.registry import set_current_agent
from koil.koil import Koil, get_current_koil
from koil.loop import koil
from fakts import Fakts, get_current_fakts, Config

try:
    from rich.traceback import install

    install()
except:
    pass
import logging


logger = logging.getLogger(__name__)


class AgentException(Exception):
    pass


class AgentConfig(Config):
    type: TransportProtocol = TransportProtocol.WEBSOCKET
    debug: bool = False
    kwargs: dict = {}

    class Config:
        group = "arkitekt.agent"


class Agent:
    def __init__(
        self,
        *args,
        auto_login=True,
        auto_connect=True,
        loop=None,
        register=True,
        with_monitor=False,
        herre: Herre = None,
        koil: Koil = None,
        fakts: Fakts = None,
        config: AgentConfig = None,
        transport_registry: TransportRegistry = None,
        strict=False,
        **kwargs,
    ) -> None:
        super().__init__(*args, **kwargs)

        self.auto_login = auto_login
        self.auto_connect = auto_connect
        self.strict = strict
        self.config = config
        self.herre = herre or get_current_herre()
        self.fakts = fakts or get_current_fakts()

        self.ward: ArkitektWard = get_ward_registry().get_ward_instance("arkitekt")
        self.monitor = Monitor("Agent") if with_monitor else None
        self.panel = self.monitor.create_agent_panel(self) if self.monitor else None
        self.message_queue = asyncio.Queue()
        self.transport: AgentTransport = None
        self.transport_registry = transport_registry or get_current_transport_registry()

        if register:
            set_current_agent(self)

    @abstractmethod
    async def handle_bounced_provide(self, bounced_provide: BouncedProvideMessage):
        raise NotImplementedError()

    @abstractmethod
    async def handle_bounced_unprovide(
        self, bounced_unprovide: BouncedUnprovideMessage
    ):
        raise NotImplementedError()

    @abstractmethod
    async def handle_bounced_assign(
        self, bounced_assign: BouncedForwardedAssignMessage
    ):
        raise NotImplementedError()

    @abstractmethod
    async def handle_bounced_unassign(
        self, bounced_unassign: BouncedForwardedUnassignMessage
    ):
        raise NotImplementedError()

    async def broadcast(self, message: MessageModel):

        logger.info(f"Provider: [green] Received message {message}")
        try:
            if isinstance(message, BouncedProvideMessage):
                logger.info("Received Provide Request")
                assert (
                    message.data.template is not None
                ), "Received Provision that had no Template???"
                await self.handle_bounced_provide(message)

            elif isinstance(message, BouncedUnprovideMessage):
                logger.info("Received Unprovide Request")
                assert (
                    message.data.provision is not None
                ), "Received Unprovision that had no Provision???"
                await self.handle_bounced_unprovide(message)

            elif isinstance(message, BouncedForwardedAssignMessage):
                logger.info("Received Assing Request")
                assert (
                    message.data.provision is not None
                ), "Received Assignment that had no Provision???"
                await self.handle_bounced_assign(message)

            elif isinstance(message, BouncedForwardedUnassignMessage):
                logger.info("Received Unassign Request")
                assert (
                    message.data.provision is not None
                ), "Received Assignment that had no Provision???"
                await self.handle_bounced_unassign(message)

            else:
                logger.info(f"Provider: [red] Received unknown message {message}")
        except Exception as e:
            logger.exception(e)

    async def on_transport_connected(self):
        return None

    async def on_transport_about_to_connect(self):
        return None

    async def on_transport_about_to_disconnect(self):
        return None

    async def aprovide(self):
        caused_ward_connect = False
        try:
            if not self.fakts.loaded:
                await self.fakts.aload()

            if not self.herre.logged_in:
                await self.herre.alogin()

            if not self.ward.connected:
                assert (
                    self.auto_connect
                ), "We have not connected to Arkitekt before and autoconnect was set to false. Please connect a ward before or set auto_connect to True"
                await self.ward.aconnect()
                caused_ward_connect = True

            if not self.config:
                self.config = await AgentConfig.from_fakts(fakts=self.fakts)

            self.transcript = self.ward.transcript
            self.transport = self.transport_registry.get_agent_transport_for_protocol(
                self.config.type
            )(self.config.kwargs, broadcast=self.broadcast)

            await self.on_transport_about_to_connect()
            await self.transport.aconnect()

        except asyncio.CancelledError as e:

            if self.transport:
                await self.on_transport_about_to_disconnect()
                await self.transport.adisconnect()
            if caused_ward_connect:
                await self.ward.adisconnect()
            raise e

        except Exception as e:
            logger.exception(e)
            raise e

    def provide(self, **kwargs):

        with get_current_console().status("Providing"):
            return koil(self.aprovide(), **kwargs)



from herre.wards.graphql import ParsedQuery


PORTS_FR = """
  args {
      __typename
      key
      required
      description
      widget {
        __typename
        dependencies
        ... on QueryWidget {
          query  
        }
        ... on SearchWidget {
          query
        }
        ... on StringWidget {
          placeholder
        }
        ... on SliderWidget {
              min
              max  
        }
      }
      label
      ... on StructureArgPort {
        identifier
      }
      ... on EnumArgPort {
        options
      }
      ... on ListArgPort {
        child {
          __typename
          key
          required
          description
          label
          ... on StructureArgPort {
            identifier
          }
        }
        
      }
  }
  kwargs {
      __typename
      key
      required
      description
      widget {
        __typename
        dependencies
        ... on QueryWidget {
          query  
        }
        ... on SearchWidget {
          query
        }
        ... on StringWidget {
          placeholder
        }
        ... on SliderWidget {
          min
          max
        }
      }
      label
      ... on StructureKwargPort {
        defaultID
        identifier
        
      }
      ... on IntKwargPort {
        defaultInt
      }
      ... on BoolKwargPort {
        defaultBool
      }
      ... on StringKwargPort {
        defaultString
      }
      ... on EnumKwargPort {
        defaultEnum
        options
      }
      ... on ListKwargPort {
        child {
          __typename
          key
          required
          description
          label
          ... on StructureKwargPort {
            identifier
          }
          ... on IntKwargPort {
            defaultInt
          }
        }
        defaultList
      }
      ... on DictKwargPort {
        child {
          __typename
          key
          required
          description
          label
          ... on StructureKwargPort {
            identifier
          }
          ... on IntKwargPort {
            defaultInt
          }
        }
        defaultDict
      }
  }
  returns {
      __typename
      key
      description
      ... on StructureReturnPort {
        identifier
      }
      ... on ListReturnPort {
        child {
          __typename
          ... on StructureReturnPort {
            identifier
          }
        }
      }
  }
"""


DETAIL_NODE_FR = (
    """
  name
  package
  interface
  id
  type
  description
  repository {
    __typename
    name
    ... on AppRepository {
      app {
        name
      }
    }
  }
"""
    + PORTS_FR
)


NODE_GET_QUERY = ParsedQuery(
    """
query Node($id: ID, $package: String, $interface: String, $template: ID, $q: String){
  node(id: $id, package: $package, interface: $interface, template: $template, q: $q){
    """
    + DETAIL_NODE_FR
    + """
  }
}
"""
)


NODE_CREATE_QUERY = ParsedQuery(
    """
mutation CreateNodeMutation(
    $description: String!,
    $args: [ArgPortInput]!,
    $kwargs: [KwargPortInput]!,
    $returns: [ReturnPortInput]!,
    $package: String!, $interface: String!,
    $name: String!
    $type: NodeTypeInput){
  createNode(description: $description, args: $args, kwargs: $kwargs, returns: $returns, package:$package, interface: $interface, name: $name, type: $type){
    """
    + DETAIL_NODE_FR
    + """
  }
}
"""
)



from typing import Optional
from pydantic.main import BaseModel
from arkitekt.messages import ReserveParams as ReserveParamsMessage
from herre.access.object import GraphQLObject


class ReserveParams(ReserveParamsMessage):
    def to_variable(self):
        return self.dict()


class TemplateParams(GraphQLObject):
    maximumInstances: Optional[int]
    maximumInstancesPerAgent: Optional[int]
